// Image classification on ResNet50

#include <iostream>
#include <fstream>
#include <string>

#include <elcorenn/elcorenn.h>
#include <opencv2/core.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>

#include "imagenet_dataclasses.h"

using namespace cv;

void preprocess_image(Mat& image) {
  Mat3f mean(224, 224, Vec3f(103.939f, 116.779f, 123.68f));

  // Preprocess opened image
  resize(image, image, Size(224, 224), 0, 0);
  image.convertTo(image, CV_32FC3);
  image -= mean;
}

int main(int argc, char** argv) {
  if (argc != 2) {
    throw std::runtime_error("Wrong number of arguments!\nUse: "
                             "./sample-resnet [image_name].jpg");
  }

  // Init elcorenn backend and load ResNet model to Elcore50
  InitBackend();

  std::string home_dir = getenv("HOME");
  std::string examples_dir = home_dir + "/elcorenn-examples-data";
  std::string model_json = examples_dir + "/models/resnet50_model.json";
  std::string model_weights = examples_dir + "/models/resnet50_weights.bin";

  auto id = LoadModel(model_json.c_str(), model_weights.c_str());

  // Open image
  auto src = imread(argv[1], IMREAD_COLOR);
  if (src.empty()) {
    throw std::ifstream::failure(
        "Couldn't open image. File or directory doesn't exist");
  }

  // Preparing data before calculations
  preprocess_image(src);
  Mat dst = Mat::zeros(1000, 1, CV_32FC1);

  // Invoke calculation of the ResNet model
  std::cout << "Invoke classification" << std::endl;
  float* inputs[1] = {src.ptr<float>(0)};
  float* outputs[1] = {dst.ptr<float>(0)};
  InvokeModel(id, inputs, outputs, 1);
  std::cout << "Finish classification" << std::endl;

  // Processing of results
  double minVal, maxVal;
  int minLoc, maxLoc;
  minMaxIdx(dst, &minVal, &maxVal, &minLoc, &maxLoc);
  std::cout << "Guessed object: " << imagenet_class.at(maxLoc);
  std::cout << " (Probability: " << maxVal << ")" << std::endl;

  return 0;
}
